

------------------------------------------------
---
--- Overload virtual class member function in Lua.
--- 
--- * Implementation of "Cross language polymorphism"
--- * Typesafe
--- 
--- Example:
---  Here drawobect() called from Lua or C++ executes only:
---   code-1: if using glviewer.h  (default case)
---   code-2: if using example.lua (Lua-overwrite case) 
---   code-3: if using user.h      (C++ (re-)overwrite case)
--- 
---  glviewer.h:
---   LuaBindBegin
---   struct glViewer  {
---     virtual void drawobject();
---     virtual ~glViewer() {};
---   };
---   LuaBindEnd
---   LuaBindVirtual(glViewer,drawobject,0)
---   
---  glviewer_c.cxx:
---   #include "ldf/swigext_virtual.h"
---   #include "ldf/swigext_lua.h"
---   SWIGEXT_DECLARE_MODULE_INFO(glviewer);
---   void glViewer::drawobject() {
---     lua_State* L = ldf_lua_state();
---     if(swigext_callvirtual(L,"gltools","glViewer","drawobject",0))
---     /* default code follows here: */
---     // code-1
---   }
---   
---  example.lua
---   function mydrawobject(self)
---     -- code-2
---     print("mydrawobject")
---     print(self)
---   end
---   swigext_virtual(gltools,"glViewer","drawobject",mydrawobject)
---   
---  user.h
---   struct my_glViewer : glViewer  {
---     virtual void drawobject() { // code-3 }
---   };

function swigext_virtual(module,clas,member,func)
  
  if type(module) == "string" then
   swigext_virtual(_G[module],clas,member,func)
   return
  end
  
  if not module then error("arg1: unknown module") end
  if type(module[clas]) ~= "function" then error("arg2: unknown class name") end
  if type(func) ~= "function" and func ~= nil then error("arg4: function or nil expected") end
  
  -- first in module
  local mt = getmetatable(module)
  if type(mt.SIWGEXT) ~= "table" then
    mt.SIWGEXT={}
    mt.SIWGEXT.virtuals={}
  end
  
  -- first in module.clas
  if type(mt.SIWGEXT.virtuals[clas]) ~= "table" then
   mt.SIWGEXT.virtuals[clas]={}
  end
  
  mt.SIWGEXT.virtuals[clas][member]=func
end



------------------------------------------------
---
--- Extend/Replace an object-member with a Lua-object.
--- 
--- * No type restrictions
--- * Lua-only
--- * replaced object-member still accessible
--- 
--- Example "extention":
--- 
---   a=da1Create(10)
---   swigext_extend(a,"mydata")
---   b=a          -- b inherits this feature from a
---   b.mydata=123
---   
---   -- temporary disable extention
---   swigext_extend_off(b)  -- changes only behavior of b
---   for i=1,10 do b[i]=i end
---   swigext_extend_on(b)
--- 
--- Example "overload":
---   
---   p=GnuPlot{}
---   
---   function mywait(gp)
---    local base_wait = swigext_base(gp,"wait")
---    print("mywait ...")
---    base_wait(gp)          -- or: swigext_basecall(gp,"wait")
---    print("... mywait")
---   end
---   
---   swigext_extend(p,"wait")
---   p.wait=mywait
---   ...
---   p:plot{x,y}
---   p:wait()
---   

function swigext_extend(o,name)
 local mt = swigext_extend_on(o)
 if not mt[name] then
  mt[name]={val=nil}
 end
end

function swigext_extend_on(o)
 local mto = getmetatable(o)
 if not mto.SWIGEXT then mto.SWIGEXT={} end
 local mt = mto.SWIGEXT
 local t = swig_type(o)
 
 -- replacement func for __index
 local function index(o,m)
  local mt = getmetatable(o).SWIGEXT
  if mt[m] then return mt[m].val end
  local t = swig_type(o)
  return mt[t].index(o,m)
 end
 
 -- replacement func for __newindex
 local function newindex(o,m,val)
  local mt = getmetatable(o).SWIGEXT
  if mt[m] then mt[m].val=val ; return end
  local t = swig_type(o)
  return mt[t].newindex(o,m,val)
 end
 
 -- save old and replace __index, __newindex
 if not mt[t] then
  mt[t]={}
  mt[t].index=mto["__index"]
  mt[t].newindex=mto["__newindex"]
  mto["__index"]=index
  mto["__newindex"]=newindex
 end
 return mt
end

function swigext_extend_off(o)
 -- restore old __index, __newindex
 local t = swig_type(o)
 local mto = getmetatable(o)
 local mt = mto.SWIGEXT
 
 if not mt[t] then return false end
 
 mto["__index"]=mt[t].index
 mto["__newindex"]=mt[t].newindex
 mt[t]=nil
 return true
end

function swigext_base(o,member)
 local off = swigext_extend_off(o)
 local base = o[member]
 if off then
  swigext_extend_on(o)
 end
 return base
end

function swigext_basecall(o,member,...)
 local f = getmetatable(o)[".fn"][member]
 return f(o,unpack(arg))
end




















